# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
"""
Generate Foucaugrams (Foucault Knife Edge images) for lens testing.

Documentation and examples are available at <https://lenstest.readthedocs.io>
"""

import numpy as np
import matplotlib.pyplot as plt
import lenstest.lenstest

__all__ = ('foucault_mask',
           'foucaugram',
           'foucault_plot'
           )

def foucault_mask(RoC, x_offset, z_offset, conic, X, Y, phi=0):
    """
    Create Foucault knife edge image with specified points.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The Ronchi grating is located at RoC + offset and oriented so lines
    are perpendicular to the x-axis

    The conic section is specified by conic:
          conic = ∞ for surface that is flat
          conic > 0 for surface that is an oblate spheroid
          conic = 0 for surface that is a sphere
          0<conic<-1 for surface that is a prolate spheroid
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        RoC: radius of curvature of mirror [mm]
        lines_per_mm: line pairs per mm [1/mm]
        offset: axial offset of grating from center of mirror's RoC [mm]
        conic: conic constant or Schwartzchild constant [-]
        X, Y: grid of points to evaluate

    Returns:
        array of points blocked by knife edge
    """
    sagitta = lenstest.lenstest.sagitta(RoC, conic, X, Y)

    # find the x value where the ray passes through the Ronchi ruling
    Xr = X * np.cos(phi)+Y*np.sin(phi)
    Lx = Xr * (-z_offset - sagitta * conic)/(RoC + sagitta * conic)

    # create mask for Knife Edge
    T_mask = Lx < x_offset

    return T_mask


def foucaugram(D, RoC, x_offset, z_offset, conic, phi=0, N=100000, invert=False, random=True):
    """
    Create Ronchigram for points on a grid or randomly selected.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The Ronchi grating is displaced from the center of the mirror by offset.
    The Ronchi rulings are oriented so lines are perpendicular to the x-axis

    The conic section is specified by conic:
          conic = ∞ for surface that is flat
          conic > 0 for surface that is an oblate spheroid
          conic = 0 for surface that is a sphere
          0<conic<-1 for surface that is a prolate spheroid
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        lines_per_mm: line pairs per mm [1/mm]
        offset: axial offset of grating from true focus [mm]
        conic: conic constant or Schwartzchild constant [-]
        N: number of points to generate
        invert: boolean to draw dark or light areas
        random: if False generate points on a grid

    Returns:
        x, y: masked listed of points to plot
    """
    X, Y = lenstest.lenstest.XY_test_points(D, N=N, random=random)

    T_mask = foucault_mask(RoC, x_offset, z_offset, conic, X, Y, phi=phi)

    if invert:
        T_mask = np.logical_not(T_mask)

    # hide all points that will appear dark
    x_image = np.ma.masked_where(T_mask, X)
    y_image = np.ma.masked_where(T_mask, Y)
    return x_image, y_image


def foucault_plot(D, RoC, x_offset, offset, conic, phi=0, init=True, figsize=(10, 5)):
    """Plot the Foucault knife edge image."""
    x, y = foucaugram(D, RoC, x_offset, offset, conic, phi=phi)

    if init:
        plt.subplots(1, 2, figsize=figsize)

    plt.subplot(1, 2, 1)
    plt.gca().set_facecolor("black")
    plt.plot(x, y, 'o', markersize=0.1, color='white')
    plt.gca().set_aspect('equal')
    lenstest.lenstest.draw_circle(D/2, color='green')
    plt.ylim(-D/2*1.2, D/2*1.2)
    plt.xlim(-D/2*1.2, D/2*1.2)
    plt.title("D=%.1fmm, RoC=%.1fmm, K=%.2f" % (D, RoC, conic))
    plt.xlabel("Mirror/Lens Plane (mm)")
    plt.ylabel("Mirror/Lens Plane (mm)")

    plt.subplot(1, 2, 2)
    x, y = lenstest.lenstest.knife_polygon(D/2, phi, x_offset)
    r_spot = abs(offset * D/2/RoC)
    lenstest.lenstest.draw_circle(r_spot, color='green')
    plt.gca().set_aspect('equal')
    plt.fill(x, y, color='black', alpha=0.8)
    plt.xlabel("Knife Edge Plane (mm)")

    size=r_spot*4
    plt.ylim(-size, size)
    plt.xlim(-size, size)

    phid = np.degrees(phi)
    plt.title('Δz=%.2fmm, Δx=%.2fmm, ϕ=%.0f°'%(offset, x_offset, phid) )



# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
"""
Generate ronchigrams for lens testing.

Documentation and examples are available at <https://lenstest.readthedocs.io>
"""

import numpy as np
import matplotlib.pyplot as plt
import lenstest

__all__ = ('ronchi_mask',
           'ronchigram',
           'ronchi_plot')

def ronchi_mask(RoC, lpm, z_offset, conic, X, Y, mask=False, phi=0):
    """
    Create Ronchigram with random points.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The Ronchi grating is located at RoC + z_offset and oriented so lines
    are perpendicular to the x-axis

    The conic section is specified by conic:
          conic = ∞ for surface that is flat
          conic > 0 for surface that is an oblate spheroid
          conic = 0 for surface that is a sphere
          0<conic<-1 for surface that is a prolate spheroid
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        lpm: line pairs per mm [1/mm]
        z_offset: axial z_offset of grating from center of mirror's RoC [mm]
        conic: conic constant or Schwartzchild constant [-]
        X, Y: grid of points to evaluate
        invert: boolean to draw dark or light areas

    Returns:
        x, y: masked listed of points to plot
    """
    if mask:
        sagitta = 0
    else:
        sagitta = lenstest.lenstest.sagitta(RoC, conic, X, Y)

    # find the x value where the ray passes through the Ronchi ruling
    Xr = X * np.cos(phi)+Y*np.sin(phi)
    Lx = Xr * (-z_offset - sagitta * conic)/(RoC + sagitta * conic)

    # create mask for Ronchi Ruling
    T = (np.abs(2*lpm*Lx)+0.5).astype(int)

    T_mask = T % 2 == 0

    return T_mask


def ronchigram(D, RoC, lpm, z_offset, conic, phi=0,\
               N=100000, invert=False, random=True, mask=False):
    """
    Create Ronchigram for points on a grid or randomly selected.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The Ronchi grating is displaced from the center of the mirror by z_offset.
    The Ronchi rulings are oriented so lines are perpendicular to the x-axis

    The conic section is specified by conic:
          conic = ∞ for surface that is flat
          conic > 0 for surface that is an oblate spheroid
          conic = 0 for surface that is a sphere
          0<conic<-1 for surface that is a prolate spheroid
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        lpm: line pairs per mm [1/mm]
        z_offset: axial z_offset of grating from true focus [mm]
        conic: conic constant or Schwartzchild constant [-]
        N: number of points to generate
        invert: boolean to draw dark or light areas
        random: if False generate points on a grid

    Returns:
        x, y: masked listed of points to plot
    """
    X, Y = lenstest.lenstest.XY_test_points(D, N=N, random=random)

    T_mask = ronchi_mask(RoC, lpm, z_offset, conic, X, Y, mask=mask, phi=phi)

    if invert:
        T_mask = np.logical_not(T_mask)

    # hide all points that should be dark
    x_mask = np.ma.masked_where(T_mask, X)
    y_mask = np.ma.masked_where(T_mask, Y)

    if mask:
        x_mask *= abs(z_offset/RoC)
        y_mask *= abs(z_offset/RoC)

    return x_mask, y_mask

def ronchi_plot(D, RoC, lp_per_mm, z_offset, conic, phi=0, init=True):
    """Plot the Surface Ronchigram."""
    x, y = ronchigram(D, RoC, lp_per_mm, z_offset, conic, phi=phi)

    if init:
        plt.subplots(1, 2, figsize=(10, 5))

    plt.subplot(1, 2, 1)
    plt.gca().set_facecolor("black")
    plt.plot(x, y, 'o', markersize=0.1, color='white')
    plt.gca().set_aspect('equal')
    lenstest.lenstest.draw_circle(D/2, color='green')
    plt.ylim(-D/2*1.2, D/2*1.2)
    plt.xlim(-D/2*1.2, D/2*1.2)
    plt.title("D=%.1fmm, RoC=%.1fmm, K=%.2f" % (D, RoC, conic))
    plt.xlabel("Mirror/Lens Plane (mm)")
    plt.ylabel("Mirror/Lens Plane (mm)")

    x, y = ronchigram(D, RoC, lp_per_mm, z_offset, conic, phi=phi, mask=True)
    plt.subplot(1, 2, 2)
    plt.gca().set_facecolor("black")
    plt.plot(x, y, 'o', markersize=0.6, color='white')
    r_spot = D/2 / RoC*z_offset
    lenstest.lenstest.draw_circle(r_spot, color='green')
    plt.gca().set_aspect('equal')
    plt.xlabel("Ronchi Ruling Plane (mm)")

    size=r_spot*1.2
    plt.ylim(-size, size)
    plt.xlim(-size, size)

    plt.title('Δz=%.2fmm, %.0f lp/mm'%(z_offset, lp_per_mm) )



# pylint: disable=invalid-name
"""
Utility routines for lens testing.

Documentation and examples are available at <https://lenstest.readthedocs.io>
"""

import numpy as np
import matplotlib.pyplot as plt

__all__ = ('XY_test_points',
           'knife_polygon',
           'circle_polygon',
           'draw_circle',
           'sagitta')

def sagitta(RoC,conic,X,Y):
    """
    Calculate sagitta for conic surface at points X,Y.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The Ronchi grating is located at RoC + z_offset and oriented so lines
    are perpendicular to the x-axis

    The conic section is specified by conic:
          conic = ∞ for surface that is flat,
          conic > 0 for surface that is an oblate spheroid,
          conic = 0 for surface that is a sphere,
          0<conic<-1 for surface that is a prolate spheroid,
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        RoC: radius of curvature of mirror [mm]
        lines_per_mm: line pairs per mm [1/mm]
        offset: axial offset of grating from center of mirror's RoC [mm]
        conic: conic constant or Schwartzchild constant [-]
        X,Y: grid of points to evaluate

    Returns:
        array of points blocked by knife edge
    """
    # find the x value where the ray passes through the Ronchi ruling
    Pr_sqr = X*X + Y*Y
    heights = Pr_sqr/(RoC+np.sqrt(RoC*RoC-(1+conic)*Pr_sqr))
    return heights

def draw_circle(R,X0=0,Y0=0,color='black'):
    """Draw a circle."""
    theta = np.linspace(0, 2*np.pi, 100)
#    plt.gca().set_aspect('equal')
    plt.plot(X0+R*np.sin(theta), Y0+R*np.cos(theta), color=color)


def XY_test_points(D, N=100000, random=True):
    """
    Generate test points for lens test.

    Args:
        D: diameter of mirror [mm]
        N: number of points to generate
        random: if False generate points on a grid
        phi: rotation of knife edge from vertical (positive ==> CCW)

    Returns:
        X,Y: arrays of test points
    """
    if random:
        U1 = np.random.uniform(size=N)
        U2 = np.random.uniform(size=N)
        X = D/2 * np.sqrt(U2) * np.cos(2 * np.pi * U1)
        Y = D/2 * np.sqrt(U2) * np.sin(2 * np.pi * U1)
    else :
        gridpts = np.linspace(-D/2, D/2, int(np.sqrt(N)))
        x_grid, y_grid = np.meshgrid(gridpts, gridpts)
        r_mask = x_grid**2+y_grid**2 > D*D/4
        X = np.ma.masked_where(r_mask, x_grid)
        Y = np.ma.masked_where(r_mask, y_grid)
    return X, Y

def knife_polygon(r,phi,dx=0):
    """Create a polygon for a knife edge."""
    r *= 1.5
    rad = phi + np.pi/2

    x=np.full(6,dx*np.cos(phi))
    y=np.full(6,dx*np.sin(phi))

    x[1] = x[0] + r*np.cos(rad)
    y[1] = y[0] + r*np.sin(rad)

    rad=rad+np.pi/2
    x[2] = x[1] + r*np.cos(rad)
    y[2] = y[1] + r*np.sin(rad)

    rad=rad+np.pi/2
    x[3] = x[2] + 2*r*np.cos(rad)
    y[3] = y[2] + 2*r*np.sin(rad)

    rad=rad+np.pi/2
    x[4] = x[3] + r*np.cos(rad)
    y[4] = y[3] + r*np.sin(rad)
    return x,y

def circle_polygon(R,X0=0,Y0=0):
    """Create a polygon for a circle."""
    theta = np.linspace(0, 2*np.pi, 100)
#    plt.gca().set_aspect('equal')
    return X0+R*np.sin(theta), Y0+R*np.cos(theta)
